/*********************************************************************************************************
* Software License Agreement (BSD License)                                                               *
* Author: Sebastien Decugis <sdecugis@nict.go.jp>							 *
*													 *
* Copyright (c) 2009, WIDE Project and NICT								 *
* All rights reserved.											 *
* 													 *
* Redistribution and use of this software in source and binary forms, with or without modification, are  *
* permitted provided that the following conditions are met:						 *
* 													 *
* * Redistributions of source code must retain the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer.										 *
*    													 *
* * Redistributions in binary form must reproduce the above 						 *
*   copyright notice, this list of conditions and the 							 *
*   following disclaimer in the documentation and/or other						 *
*   materials provided with the distribution.								 *
* 													 *
* * Neither the name of the WIDE Project or NICT nor the 						 *
*   names of its contributors may be used to endorse or 						 *
*   promote products derived from this software without 						 *
*   specific prior written permission of WIDE Project and 						 *
*   NICT.												 *
* 													 *
* THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED *
* WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A *
* PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR *
* ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT 	 *
* LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS 	 *
* INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR *
* TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF   *
* ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.								 *
*********************************************************************************************************/

/* Create and send a message, and receive it */

/* Note that we use both sessions and the argument to answer callback to pass the same value.
 * This is just for the purpose of checking everything went OK.
 */

#include "app_test.h"

static sess_reg_t * atst_cli_reg = NULL;

typedef struct {
	int32_t		randval;	/* a random value to store in Test-AVP */
	struct timespec ts;		/* Time of sending the message */
} atst_mess_info_t;

/* Cb called when an answer is received */
static void atst_cb_ans(void * data, msg_t ** msg)
{
	atst_mess_info_t * mi = NULL;
	struct timespec ts;
	
	CHECK_VOID( clock_gettime(CLOCK_REALTIME, &ts) );

	/* Search the session, retrieve its data */
	{
		sess_id_t *sess = NULL;
		msg_avp_t * src = NULL;
		msg_avp_data_t * avpdata = NULL;
		int new;
		
		CHECK_VOID( msg_search_avp(*msg, atst_sess_id, &src) );
		CHECK_VOID( msg_avp_data( src, &avpdata ) );
		
		CHECK_VOID( sess_fromsid( (char *)avpdata->avp_data->os.data, avpdata->avp_data->os.len, &sess, &new) );
		ASSERT( new == 0 );
		
		CHECK_VOID( sess_data_dereg( sess, atst_cli_reg, (void *)&mi) );
		
		TRACE_DEBUG( INFO, "%p %p", mi, data);
		
		ASSERT( (void *)mi == data );
		
		CHECK_VOID( sess_unlink( &sess ) );
	}
	
	/* Now log content of the answer */
	fprintf(stderr, "RECV ");
	
	/* Value of Test-AVP */
	{
		msg_avp_t * src = NULL;
		msg_avp_data_t * avpdata = NULL;
		
		CHECK_VOID( msg_search_avp(*msg, atst_avp, &src) );
		if (src) {
			CHECK_VOID( msg_avp_data( src, &avpdata ) );
			fprintf(stderr, "%x (%s) ", avpdata->avp_data->i32, (avpdata->avp_data->i32 == mi->randval) ? "Ok" : "PROBLEM");
		} else {
			fprintf(stderr, "no_Test-AVP ");
		}
	}
	
	/* Value of Result Code */
	{
		msg_avp_t * src = NULL;
		msg_avp_data_t * avpdata = NULL;
		
		CHECK_VOID( msg_search_avp(*msg, atst_res_code, &src) );
		if (src) {
			CHECK_VOID( msg_avp_data( src, &avpdata ) );
			fprintf(stderr, "Status: %d ", avpdata->avp_data->i32);
		} else {
			fprintf(stderr, "no_Result-Code ");
		}
	}
	
	/* Value of Origin-Host */
	{
		msg_avp_t * src = NULL;
		msg_avp_data_t * avpdata = NULL;
		
		CHECK_VOID( msg_search_avp(*msg, atst_origin_host, &src) );
		if (src) {
			char * str;
			CHECK_VOID( msg_avp_data( src, &avpdata ) );
			str = malloc(avpdata->avp_data->os.len + 1);
			memcpy(str, avpdata->avp_data->os.data, avpdata->avp_data->os.len);
			str[avpdata->avp_data->os.len] = '\0';
			fprintf(stderr, "From '%s' ", str);
			free(str);
		} else {
			fprintf(stderr, "no_Origin-Host ");
		}
	}
	
	/* Value of Origin-Realm */
	{
		msg_avp_t * src = NULL;
		msg_avp_data_t * avpdata = NULL;
		
		CHECK_VOID( msg_search_avp(*msg, atst_origin_realm, &src) );
		if (src) {
			char * str;
			CHECK_VOID( msg_avp_data( src, &avpdata ) );
			str = malloc(avpdata->avp_data->os.len + 1);
			memcpy(str, avpdata->avp_data->os.data, avpdata->avp_data->os.len);
			str[avpdata->avp_data->os.len] = '\0';
			fprintf(stderr, "('%s') ", str);
			free(str);
		} else {
			fprintf(stderr, "no_Origin-Realm ");
		}
	}
	
	/* Now compute how long it took */
	if (ts.tv_nsec > mi->ts.tv_nsec) {
		fprintf(stderr, "in %d.%06ld sec", 
				(int)(ts.tv_sec - mi->ts.tv_sec),
				(long)(ts.tv_nsec - mi->ts.tv_nsec) / 1000);
	} else {
		fprintf(stderr, "in %d.%06ld sec", 
				(int)(ts.tv_sec + 1 - mi->ts.tv_sec),
				(long)(1000000000 + ts.tv_nsec - mi->ts.tv_nsec) / 1000);
	}
	
	fprintf(stderr, "\n");
	fflush(stderr);
	
	/* Free the message */
	CHECK_VOID(msg_free(*msg, 1));
	*msg = NULL;
	
	free(mi);
	
	return;
}

static void atst_cli_test_message(void)
{
	msg_t * req = NULL;
	atst_mess_info_t * mi = NULL;
	sess_id_t *sess = NULL;
	
	TRACE_DEBUG(FULL, "Creating a new message for sending.");
	
	/* Create the request from template */
	CHECK_VOID( msg_new( atst_cmd_r, MSGFL_FROM_TEMPLATE | MSGFL_ALLOW_ETEID, &req ) );
	
	/* Create a new session */
	CHECK_VOID( sess_new( &sess, SESSION_NEW_DEFAULT, "app_test" ) );
	
	/* Create the random value to store with the session */
	mi = malloc(sizeof(atst_mess_info_t));
	if (mi == NULL) {
		log_error("malloc failed: %s", strerror(errno));
		return;
	}
	
	mi->randval = (int32_t)random();
	
	/* Now set all AVPs values */
	
	/* Session-Id */
	{
		msg_avp_t * dst = NULL;
		avp_value_t data;
		
		CHECK_VOID( msg_search_avp(req, atst_sess_id, &dst) );
		
		CHECK_VOID( sess_getsid(sess, (void *)&data.os.data) );
		data.os.len  = strlen((char *)data.os.data);
		
		CHECK_VOID( msg_avp_setvalue( dst, &data) );
	}
	
	/* Set the Destination-Realm AVP */
	{
		msg_avp_t * dst = NULL;
		avp_value_t data;
		
		data.os.data = (unsigned char *)(atst_conf->dest_realm);
		data.os.len  = strlen(atst_conf->dest_realm);
		
		CHECK_VOID( msg_search_avp(req, atst_dest_realm, &dst) );
		CHECK_VOID( msg_avp_setvalue( dst, &data) );
	}
	
	/* Set or delete the Destination-Host AVP */
	{
		msg_avp_t * dst = NULL;
		avp_value_t data;
		
		CHECK_VOID( msg_search_avp(req, atst_dest_host, &dst) );
		
		if (atst_conf->dest_host) {
			data.os.data = (unsigned char *)(atst_conf->dest_host);
			data.os.len  = strlen(atst_conf->dest_host);

			CHECK_VOID( msg_avp_setvalue( dst, &data) );
		} else {
			CHECK_VOID( msg_free ( dst, 0 ) );
		}
	}
	
	/* Set the Origin-Host AVP */
	{
		msg_avp_t * dst = NULL;
		avp_value_t data;
		
		data.os.data = (unsigned char *)(g_pconf->diameter_identity);
		data.os.len  = strlen(g_pconf->diameter_identity);
		
		CHECK_VOID( msg_search_avp(req, atst_origin_host, &dst) );
		CHECK_VOID( msg_avp_setvalue( dst, &data) );
	}
	
	/* Set the Origin-Realm AVP */
	{
		msg_avp_t * dst = NULL;
		avp_value_t data;
		
		data.os.data = (unsigned char *)(g_pconf->diameter_realm);
		data.os.len  = strlen(g_pconf->diameter_realm);
		
		CHECK_VOID( msg_search_avp(req, atst_origin_realm, &dst) );
		CHECK_VOID( msg_avp_setvalue( dst, &data) );
	}
	
	/* Set the Test-AVP AVP */
	{
		msg_avp_t * dst = NULL;
		avp_value_t data;
		
		data.i32 = mi->randval;
		
		CHECK_VOID( msg_search_avp(req, atst_avp, &dst) );
		CHECK_VOID( msg_avp_setvalue( dst, &data) );
	}
	
	CHECK_VOID( clock_gettime(CLOCK_REALTIME, &mi->ts) );
	
	/* Store this value in the session */
	CHECK_VOID( sess_data_reg ( sess, atst_cli_reg, (void *) mi, (void (*)(void *))free ) ); 
	
	/* Log sending the message */
	fprintf(stderr, "SEND %x to '%s' (%s)\n", mi->randval, atst_conf->dest_realm, atst_conf->dest_host?:"-" );
	fflush(stderr);
	
	/* Send the request */
	CHECK_VOID( msg_send( &req, atst_cb_ans, (void *)mi ) );
	
	return;
}

int atst_cli_init(void)
{
	CHECK( sess_regext(&atst_cli_reg) );
	
	CHECK( atst_sig_init(atst_cli_test_message) );
	
	return 0;
}

void atst_cli_fini(void)
{
	(void) sess_deregext(atst_cli_reg);
	
	atst_sig_fini();
	
	return;
};
